#!/usr/bin/perl -w
# Generate _keywords.index & _commands.index - used by MicroEmacs
#
#       /\
#      /  \             (C) Copyright 2019 Parliament Hill Computers Ltd.
#      \  /             All rights reserved.
#       \/
#        .              Author: Alain Williams, <addw@phcomp.co.uk> March 2019
#        .
#         .             SCCS: @(#)	1.6 06/20/22 23:58:08
#           .
# 
# This is free software licensed under the GPL, version 2 or later:
#    http://opensource.org/licenses/gpl-license
#
# The purpose is to generate the various index files that Emacs uses
# to find its way round help files. It will also do some syntax checking.
#
# To understand the format of the help files see: help-on-help.help
#
# To use this run it in the directory that contains various .help or
# .mac files - but not both.
# It will recreate the files _keywords.index & _commands.index
# It takes some command line options.
#
# This is work in progress. It works but does not do everything that it could do.

# TODO
# Parse searches & warn if the keys are not found.
# Note that this can only be a warning that the keys are not in the current
# directory, the help file search path can-be/is user specific.
# Eg: ||>search: bilbo baggins ||
# Check if all files are referenced somewhere.

# Ought to check **v:variable** - syntax is OK. But what to do if it is wrong ?
# With other things that are almost some syntax - they are not recognised.

use strict;

# Change some defaults:
# no_ignore_case	options to be case sensitive
# bundling		single letter options can be joined, eg -ab
# no_auto_abbrev	long options must be given in full
# require_order		options cannot be put after first non option
# auto_version		--version prints $VERSION
use Getopt::Long qw(:config no_ignore_case bundling no_auto_abbrev require_order auto_version);

my ($progname) = $0 =~ m|([^/]+)$|;
my $Verbose = 0;			# True if to be chatty
my $Trace = 0;				# Trace parsed input as it is read -- debug option
my $Quiet = 0;				# Suppress non error messages
my $QuietMissingReference = 0; # Do not warn about references to files that have not been seen
my $Test = 0;               # Test mode - this does what I want today
my $VERSION = "1.6 06/20/22";		# Getopt::Long prints this when it sees --version
my $ErrCnt = 0;				# Count of errors
my $MaxErrors = 100;
my $Now = time;
my $ProcAll = 0;			# Ignore **!stopindex directives
my $DoDump = 0;             # Dump to $dumpFile
my $dumpFile = "found.dump";# Name of --dump file

# Index all the keys here
# Key is command name
# Value is hash:
#    file       Name of file where it was seen
#    line       Line number in file
#    index      Name of file with sufficies removed
my %cmds;

# Something about every file found.
# Key is the file name with .gz and .help removed
# Also remove any language_territory (eg en_GB) - this ought to be used at some point
# Value is a hash with keys::
#    title	File description as per **t:title/description**
#    keys	Array of keys
# Eg: $files{faq}{title} = 'FAQ Frequently Asked Questions'
my %files;

# The names of files as they appear in a link, ie with .help, whatever
my @fileBasenames = ();

# Key is the name of a file referenced in something like ||some-link: keywords||
# The value is the number of times that the reference was seen.
# The purpose is to be able to check that every reference is to a file that has been seen,
# do this by checking to see if something exists in %files.
# This is actually done by a different way
my %filesReferenced;

# Array of all links found.
# Each element is a hash:
#   ref   Where the reference is found, ie where we saw the ||some-link: keywords||
#   file  The file of the link, eg 'some-link'. May be empty
#   flag  String of flags, eg '&='
#   kw    Keywords
my @links = ();

my @srchs = ();     # All searches

my $cnt = 0;        # How many commands seen
my $linkCnt = 0;    # How many links seen
my $WebLinkCnt = 0; # How many web links seen
my $fileCnt = 0;    # How many files seen
my $lineCnt = 0;    # How many lines seen
my $placeCnt = 0;   # How many places seen

my $seenHelp = 0;   # How many .help files seen
my $seenMac  = 0;   # How many .mac  files seen
my $inHelp;         # True if what we are reading is help text - as opposed to (probably) macro stuff

# TODO
# complain about multiple spaces, eg: **:several   words** ||:several   words||
# Check that every reference (||xxx: keys||) has a corresponding place (**: keys**)

use IO::File;
use IO::Pipe;
use Cwd;

# Print an error message, if more than 100, die:
sub Err {
	if(++$ErrCnt > $MaxErrors) {
		$ErrCnt = 0;
		ErrDie(@_);
	}

	print STDERR "$_[0]\n";
}

# Error message, try to archive input files, die
sub ErrDie {
	print STDERR "$_[0]\n";
	exit 2;
}

sub Trace {
	return unless($Trace);

	print STDERR "$_[0]\n";
}

sub Warn {
	return if($Quiet);

	print STDERR "$_[0]\n";
}

sub VerbWarn {
	return unless($Verbose);

	print STDERR "$_[0]\n";
}

# Print help/usage:
sub Help {
	print "Generates MicroEmacs help index files from the .help files in the current directory\n";
	print "Usage: $progname [options]\n";
	print "Options:\n";
	print "--dump\tWrite list of places & labels found to $dumpFile\n";
	print "--process-all\n";
	print "\tIndex all files including those that contain **!stopindex\n";
	print "-q\tQuiet - non error messages are suppressed\n";
    print "-r\tDo not complain about references to files that have not been seen\n";
    print "\tor files seen but not referenced.\n";
    print "\tThey, or references, could exist but in another directory - unseen by this program\n";
	print "-t\tTrace parsed input in some way -- debug option\n";
	print "-T\tTest option -- what this does is not defined\n";
	print "--verbose\n";
	print "-v\tVerbose\n";
	print "--help\n";
	print "-x\tExplain\n";
	print "Version: 1.6 06/20/22\n";
	exit 0;
}

# Parse options:
# Need an option that controls how to deal with @ flag: silent, complain
GetOptions(
    "dump"          => \$DoDump,
    "help|?|x"      => \&Help,
    "process-all"   => \$ProcAll,
    "q"             => \$Quiet,
    "r"             => \$QuietMissingReference,
    "t"             => \$Trace,
    "T"             => \$Test,
    "verbose|v"     => \$Verbose,   # flag
    ) || Help();


# Chase down the link and see if it exists
# Argument is a reference to a place in @links
my $once = 0;
sub FindLink {
    return if $once > 10 && $Test;
    $once++;

    my $e = $_;
#    print "ref=$e->{ref} file=$e->{file} flag='$e->{flag}' kw='$e->{kw}'\n";

    my $cmd         = index($e->{flag}, '^') != -1;  # Looking for a command
    if($cmd) {
        # Test for commands done elsewhere
        Err("$e->{ref} command does not exist: '$e->{kw}'") ;#unless(exists $cmds{$e->{kw}});
        return;
    }

    my $notComplain = index($e->{flag}, '@') != -1;  # Don't complain if not found
    my $all         = index($e->{flag}, '&') != -1;  # all keywords must match
    my $full        = index($e->{flag}, '=') != -1;  # keywords full match

    # Get an array of files to search for the keyword.
    # Use the file specified, if not use the one in ref - ie the file that it was found in
    my $lookFile;
    if($e->{file} ne "") {
        $lookFile = $e->{file};
    } else {
        ($lookFile) = $e->{ref} =~ /^([^.]+)/;
    }

    if( !exists $files{$lookFile}) {
        # This can happen when we are testing and only looking at one file
        return if($Test);

        # Otherwise it is an error
        Err("$e->{ref} refers to a file that does not exist: $e->{file}") unless $QuietMissingReference;

        return;
    }

    # Do not complain about next/previous place links.
    # TODO check that there is indeed a place to go to
    return if($e->{kw} eq "");

    # Keys in that file:
    my @keys = @{$files{$lookFile}{keys}};

    # Look through every keys string in the file
    for my $key (@keys) {
        $key = lc $key;
        my %keywords = map { $_ => 1 } split(' ', $key) if($full);

        # See if the keywords match this
        my $match = 0;     # At least one matched
        my $notmatch = 0;  # At least one did not match

        for my $kw ( split(' ', $e->{kw}) ) {
            $kw = lc $kw;
            my $found;      # Set true if this keyword matches
            if($full) {
                $found = exists $keywords{$kw};
            } else {
                # Is the key in the link somewhere in this file entry ?
                $found = index($key, $kw) != -1;
            }

            # Is one match enough ?
            return if($found && !$all);

            # Record what happened
            $match    = 1 if(  $found);
            $notmatch = 1 if( !$found);

        }

        # Matched something and no not-matches then we have at least one of what we want
        return if($all && $notmatch == 0 && $match == 1);
    }

    # If we get here then we did not match the keys in the link
    Err("$e->{ref}: link not found '$e->{file}:$e->{flag} $e->{kw}'");

#    print "$_\n" for (@keys);
#    print "******\n\n\n";

# print "file $_\n" for (sort keys %files);
#        $files{$filenx}{keys} = \@keys;

}

# Get all the .help and .help.gz files in the current directory:
my @files2process;
if($Test) {
    @files2process = ('help-on-help.help');    # For testing
} else {
    @files2process = glob("[!_]*.help [!_]*.help.gz [!_]*.mac [!_]*.mac.gz");
}

# print "$_\n" for @files2process;

# Look at all of the .help files found, scan them for keywords:
for my $file (@files2process) {
        my $fh;

        my @keys = ();
        my $fileKeywords = '';

        my $filenx = $file;     # File name as indexed, without suffix

        # Open the file. If gzipped - uncompress
        if($file =~ /\.gz$/) {
                $fh = IO::Pipe->new;
                $fh->reader('gzip', '-dc', $file);
                Warn("opened pipe from $file");
                $filenx =~ s/\.gz$//;
        } else {
                $fh = IO::File->new($file, "r") ||
                        ErrDie("Cannot open $file as: $!");
                Warn("Opened $file");
        }

        $fileCnt++;

        $filenx =~ s/\.(help|mac)$//;
        if($1 eq "help") {
            $seenHelp++;
            $inHelp = 1;
        }
        if($1 eq "mac") {
            $seenMac++;
            $inHelp = 0;
        }

        # Need to also remove language_territory
        $filenx =~ s/\.[a-z][a-z](_[A-Z][A-Z])?$//;

        push @fileBasenames, $filenx;

        # Check for bad file names:
        Err("Illegal file name: '$filenx'") unless($filenx =~ /^[a-zA-Z0-9_][-a-zA-Z0-9_]*$/);

        # These could be confused with URLs:
        Err("File name cannot be used: '$filenx'")
            if($filenx eq "http" or $filenx eq "https");

        # Read every line:
        while(<$fh>) {
                $lineCnt++;

                next if /^\*\*#/;   # Comments

                # Directive lines
                if(/^\*\*!/) {
                    my $understood = 0;
                    chop;

                    # Stop processing ?
                    if(/^\*\*!stopindex\s*(.*)/) {
                        warn "Unexpected text after **!stopindex on line $.\n" if($1 ne "");
                        last if(!$ProcAll);
                        $understood = 1;
                    }

                    if(/^\*\*!warn\s+(.*)$/) {
                        warn "Expecting text after **!warn on line $.\n" if($1 eq "");
                        warn "$1\n" if(!$Quiet);
                        $understood = 1;
                    }

                    # Ignore for help purposes:
                    if(/^\*\*!help-end\s*(.*)/) {
                        warn "Unexpected text after **!help-end on line $.\n" if($1 ne "");
                        $inHelp = 0;
                        $understood = 1;
                    }

                    # Notice for help purposes:
                    if(/^\*\*!help-begin\s*(.*)/) {
                        warn "Unexpected text after **!help-begin on line $.\n" if($1 ne "");
                        $inHelp = 1;
                        $understood = 1;
                    }

                    warn "Directive not recognised on line $. '$_'\n" unless($understood);

                    next;
                }

                # Ignore stuff in **!help-end regions
                next unless($inHelp);

                my $line = $_;  # We'll need this later

                # Process **^emacs-command**
                # Allow keywords after the command name, eg: **^emacs-command some keys **
                while(s/\*\*\^\s*([-a-z\d]+)(\s+([^*|]+))?\s*(\*\*)//) {
                        $cnt++;
                        $cmds{$1} = {} unless(exists($cmds{$1}));
                        if(exists($cmds{$1}{file})) {
                                Warn("WARNING: Command $1 seen at $file:$. first seen at $cmds{$1}{file}:$cmds{$1}{line}");
                        } else {
                                # New entry
                                $cmds{$1}{index} = $filenx;
                                $cmds{$1}{file}  = $file;
                                $cmds{$1}{line}  = $.;
                        }

                        my $keys = $1;

                        # print "Command: $1\n";
                        if(defined $3) {
                            $keys = "$1 $3";
                            $keys =~ s/^ +//;
                            $keys =~ s/ +$//;
                            $keys =~ s/ +/ /g;
                            # print "1= '$1' 3='$3' keys='$keys'\n";
                        }

                        push @keys, ("^" . $keys);
                        $placeCnt++;

                        # print "seen $1 at $.\n";
                }

                $_ = $line;
                # Process ||^emacs-command||
                # The purpose is to get a list of Emacs commands, check all have a definition
                while(s/\|\|\^\s*([-a-z\d]+)\s*(?=\|\|)//) {
                        $cmds{$1} = {} unless(exists($cmds{$1}));
                        # Only record first use:
                        unless(exists($cmds{$1}{used})) {
                                $cmds{$1}{used}   = $file;
                                $cmds{$1}{usedln} = $.;
                        }
                        $linkCnt++;
                }

                $_ = $line;
                # Process binding display **b:emacs-command;length**
                # The purpose is to get a list of Emacs commands, check all have a definition
                while(s/\*\*b:-?([-a-z\d]+)(;\d+\s*)?(?=\*\*)//) {
                        $cmds{$1} = {} unless(exists($cmds{$1}));
                        # Only record first use:
                        unless(exists($cmds{$1}{used})) {
                                $cmds{$1}{used}   = $file;
                                $cmds{$1}{usedln} = $.;
                        }
                }

                $_ = $line;
                # This will match something like ||:^ command || - should not be found, but has been found
                if(/\|\|:\^\s*([^|]*)\|\|/) {
                    Err("$file:$. Syntax error: ^ after : with command $1");
                }

                # Process ||http://URL || - check for space before ending ||
                while(s@\|\|https?://[^|\s]+(\s*)\|\|@@) {
                    Err("URL should end in a space. $file:$.")
                        if(length($1) == 0);
                    $WebLinkCnt++;
                }

                # Process ||file-name|| - later check that the file exists
                # Process ||file-name: keywords|| - later check that the file & keys exists
                # Process ||: keywords|| - later check that the keys exist in the current file
                # After a ':' can have [-&=@] flags
                # print "\nline=$_";
                while(s/\|\|([-a-z\d+]*)\s*(:-{0,2}([-&=@]*)\s*([^|]*))?(?=\|\|)//i) {
                    my $f = $1;
                    $f = '' unless defined $1;
                    $filesReferenced{$f}++ if $f ne '';
                    my $flag = $3;
                    $flag = '' unless defined $3;
                    my $kw = $4;
                    $kw = '' unless defined $4;
                    $kw =~ s/^ +//;
                    $kw =~ s/ +/ /g;
                    $kw =~ s/ +$//;
                    Err("Cannot have flag @ if file-name is empty. $file:$.")
                        if($f eq "" && index($flag, '@') != -1);
                    # print "proc || f='$f' flag='$flag' keywords='$kw'\n";
                    push @links, {
                        ref => "$file:$.",
                        file => $f,
                        flag => $flag,
                        kw => $kw
                    };
                    $linkCnt++;
                }

                $_ = $line;
                # Process ||>s:word1 word2|| & synonym ||>search:word1 word2||
                # The purpose is to check if the key words exist
                while(s/\|\|>(s|search):([&=^@]*)\s*([^|]*)\s*(?=\|\|)//) {
                        my %srch = ();
                        $srch{cmd} = $1;
                        $srch{flag} = $2;
                        $srch{kw} = $3;
                        $srch{ref} = "$file:$.";
                        $srch{file} = $filenx;
                        $srch{line} = $.;

                        # Not interested in search that prompts the user
                        push @srchs, \%srch if($3 ne "");
                }

                $_ = $line;
                # Process **:word1 word2**
                # Keywords starting '-' are ignored for indexing purposes
                # TODO they are not ignored - put into @keys for FindLink()
                while(s/\*\*:(-*)?([^*|]+)\*\*//) {
                        my $keys = $2;
                        $keys =~ s/^ +//;
                        $keys =~ s/ +/ /g;
                        $keys =~ s/ +$//;
                        push @keys, $keys;
                        $placeCnt++;
                }

                $_ = $line;
#                # Process **k: - keywords for help file
#                if(/\*\*k:(((?!\*\*).)+)\*\*/) {
#                        my $keys = $1;
#                        $keys =~ s/^ +//;
#                        $keys =~ s/ +$//;
#                        my @words = split(/ +/, $keys);
#                        $files{$filenx}{filekeys} = join("\0", @words);
#                }

                # Process **t: - title/description for help file
                if(/\*\*t:(((?!\*\*).)+?)\*\*/) {
                        my $title = $1;
                        $title =~ s/^ +//;
                        $title =~ s/ +$//;
                        $files{$filenx}{title} = $title;
                }
        }

        $files{$filenx}{keys} = \@keys;

        $fh->close;
}

Warn("Seen: $fileCnt files, $lineCnt lines, $cnt commands, $placeCnt places, $linkCnt links, $WebLinkCnt Web URLs");

Warn("WARNING: seen both .help and .mac files. Should be one of the other in a directory. ($seenHelp .help $seenMac .mac)") if($seenHelp != 0 && $seenMac != 0);


# Print all the files that we saw:
# print "file $_\n" for (sort keys %files);

# Debug - dump what we gleaned about these help files
if($DoDump) {
    my $dump = IO::File->new($dumpFile, "w") ||
            ErrDie("Cannot open/write $dumpFile as: $!");

    print $dump "Files:\n";
    print $dump "\t$_\n" for (sort keys %files);

    print $dump "Links:\n";
    my $lf = '';
    for my $l (@links) {
        my ($fn, $ln) = $l->{ref} =~ /^(.*):(.*)/;
        if($lf ne $fn) {
            print $dump "\t$fn\n";
            $lf = $fn;
        }
        print $dump "\t\t$ln\t||$l->{file}:$l->{flag} $l->{kw}||\n"
    }

    print $dump "Places:\n";
    for my $fn (sort keys %files) {
        print $dump "\t$fn\n";
        for my $kr (@{$files{$fn}{keys}}) {
            print $dump "\t\t'$kr'\n";
        }
    }

    print $dump "Commands:\n";
    for my $cmd (sort keys %cmds) {
        my $loc = '';
        $loc = "$cmds{$cmd}{file}:$cmds{$cmd}{line}" if(exists($cmds{$cmd}{index}));

        print $dump "\t$loc\t^$cmd\n";
    }

    print $dump "Searches:\n";
    $lf = '';
    for my $srch ( @srchs ) {
        print $dump "\t$srch->{file}\n" if($srch->{file} ne $lf);
        $lf = $srch->{file};
        print $dump "\t\t$srch->{line}\t$srch->{flag}\t$srch->{kw}\n";
    }

    $dump->close;
}

# Print out the names of files that we found but were not referenced,
# we did not see the file in something like ||some-file: keys ||
unless($QuietMissingReference) {
    for my $file (@fileBasenames) {
        Warn("WARNING file $file is not referenced") unless exists $filesReferenced{$file};
    }
}

# Now go through the links that we saw and see if we can find them
for (@links) {
    #    print "ref=$_->{ref} file=$_->{file} flag='$_->{flag}' kw='$_->{kw}'\n";
    FindLink($_);
}

if($seenHelp > 0) {
    # Generate the index of Emacs commands
    # This is ONLY done in a directory that contain .help files
    my $cf = IO::File->new("_commands.index", "w") ||
            ErrDie("Cannot open/write _commands.index as: $!");

    print $cf "# MicroEmacs command index file\n";
    print $cf "# **** DO NOT HAND EDIT ****\n";
    print $cf "# This file was generated by $progname\n";
    print $cf "#\n";
    for my $cmd (sort keys %cmds) {
            print $cf "c\0$cmd\0$cmds{$cmd}{index}\0\n" if(exists($cmds{$cmd}{index}));

            # While we are about this, do a couple of checks about commands
            # being defined and not used.
            Warn("WARNING: command $cmd not used, defined $cmds{$cmd}{file}:$cmds{$cmd}{line}") if(exists($cmds{$cmd}{index}) && !exists($cmds{$cmd}{used}));
            Warn("WARNING: command $cmd not defined, used $cmds{$cmd}{used}:$cmds{$cmd}{usedln}") if(!exists($cmds{$cmd}{index}) && exists($cmds{$cmd}{used}));
    }

    $cf->close;
}

# Generate the keywords index
# Each Emacs command is also one entry in this
my $kf = IO::File->new("_keywords.index", "w") ||
        ErrDie("Cannot open/write _keywords.index as: $!");

print $kf "# MicroEmacs keyword index file\n";
print $kf "# **** DO NOT HAND EDIT ****\n";
print $kf "# This file was generated by $progname\n";
print $kf "#\n";

for my $file (keys %files) {
        print $kf "f\0$file\0\n";

        print $kf "t\0$files{$file}{title}\0\n" if(exists($files{$file}{title}));

        my $keys = $files{$file}{keys};
        for my $kw (@$keys) {
                # If the keyword start # it is a command:
                if(substr($kw, 0, 1) eq '^') {
                    print $kf "c\0" . substr($kw, 1) . "\0\n";
                } else {
                    print $kf "k\0$kw\0\n";
                }
        }
}

$kf->close;

# Check all the searches
# Read the index files that we have just written, this is so that we can
# also look at index files in other directories -- at some stage
for ( @srchs ) {
    if(index($_->{flag}, '@') != -1) {
        # command name
#        FindCommand($srch->{keys}, $srch->{ref});
    } else {
        # keyword search
# Findlink looks at $_->{file} - this should search everywhere
#        FindLink();
    }
#   ref   Where the reference is found, ie where we saw the ||some-link: keywords||
#   file  The file of the link, eg 'some-link'. May be empty
#   flag  String of flags, eg '&='
#   kw    Keywords
}

exit $ErrCnt == 0 ? 0 : 1;
